#include <iostream>
#include <sstream>
#include <cstdio>
#include <vector>
#include <cmath>
using namespace std;

static const int M=40;
static const int N1=40;
static const double h=0.093;

struct point {
  double x,y,z; // position
  double nx,ny,nz; // normals (used later)
  point(double u=0, double v=0, double w=0, double nu=0, double nv=0, double nw=0)
    { x=u; y=v; z=w; nx=nu; ny=nv; nz=nw; }
  point normal() {
    return point(nx,ny,nz);
  }
};
vector<point> vertices;
vector<vector<int> > faces;

point operator+(const point &a, const point &b) {
  return point(a.x+b.x,a.y+b.y,a.z+b.z,a.nx+b.nx,a.ny+b.ny,a.nz+b.nz);
}

point operator*(const double &t, const point &b) {
  return point(t*b.x,t*b.y,t*b.z,t*b.nx,t*b.ny,t*b.nz);
}

point operator*(const point &b, const double &t) {
  return point(t*b.x,t*b.y,t*b.z,t*b.nx,t*b.ny,t*b.nz);
}

point cross(const point& a, const point& b) {
  return point(a.y*b.z-a.z*b.y,a.z*b.x-a.x*b.z,a.x*b.y-a.y*b.x);
}

double norm(const point &p) {
  return sqrt(p.x*p.x+p.y*p.y+p.z*p.z);
}

double dot(const point &a, const point &b) {
  return a.x*b.x+a.y*b.y+a.z*b.z;
}

point turn_z(point &p, double cos_angle, double sin_angle) {
  return point(cos_angle*p.x-sin_angle*p.y,sin_angle*p.x+cos_angle*p.y,p.z
              ,cos_angle*p.nx-sin_angle*p.ny,sin_angle*p.nx+cos_angle*p.ny,p.nz);
}
point turn_x(point &p, double cos_angle, double sin_angle) {
  return point(p.x,cos_angle*p.y-sin_angle*p.z,sin_angle*p.y+cos_angle*p.z
              ,p.nx,cos_angle*p.ny-sin_angle*p.nz,sin_angle*p.ny+cos_angle*p.nz);
}
point turn_y(point &p, double cos_angle, double sin_angle) {
  return point(cos_angle*p.z-sin_angle*p.x,p.y,sin_angle*p.z+cos_angle*p.x
              ,cos_angle*p.nz-sin_angle*p.nx,p.ny,sin_angle*p.nz+cos_angle*p.nx);
}

int idx;

vector<int> merg;

void add_point(double x, double y, double z, double nx=0, double ny=0, double nz=0) {
  idx++;
  vertices.push_back(point(x,y,z,nx,ny,nz));
  merg.push_back(-1);
}

void add_point(point &p) {
  idx++;
  vertices.push_back(p);
  merg.push_back(-1);
}


void mark(vector<int> &l) {
  l.push_back(idx);
}

void turn(vector<int> &src, vector<int> &dest, int k) {
  if(k==1)
    for(vector<int>::iterator it=src.begin(); it<src.end(); it++) {
      point &v=vertices[*it];
      add_point(v.z,v.x,v.y,v.nz,v.nx,v.ny);
      mark(dest);
    }
  else
    for(vector<int>::iterator it=src.begin(); it<src.end(); it++) {
      point &v=vertices[*it];
      add_point(v.y,v.z,v.x,v.ny,v.nz,v.nx);
      mark(dest);
    }
}

void push_array(vector<int> indices, int ni, int nj, vector<int> *f=NULL) {
  for(int i=0; i<ni; i++) {
    for(int j=0; j<nj; j++) {
      vector<int> v;
      v.push_back(indices[    i*(nj+1)+j]);
      v.push_back(indices[(i+1)*(nj+1)+j]);
      v.push_back(indices[(i+1)*(nj+1)+j+1]);
      v.push_back(indices[    i*(nj+1)+j+1]);
      if(f) f->push_back(faces.size());
      faces.push_back(v);
    }
  }
}

void build_surface() {
  double th,phi;
  idx=-1;
  
  vector<int> cyl_z_up; 
  for(int i=0; i<=3*M; i++) {
    th=(M_PI/2)*i/(double) M;
    add_point(1-cos(th),-1-sin(th),1,-cos(th),-sin(th),0);
    mark(cyl_z_up);
  }
  vector<int> cyl_z_down; 
  for(int i=0; i<=3*M; i++) {
    th=(M_PI/2)*i/(double) M;
    add_point(1-cos(th),-1-sin(th),-1,-cos(th),-sin(th),0);
    mark(cyl_z_down);
  }
  vector<int> cyl_x_up;
  vector<int> cyl_x_down;
  vector<int> cyl_y_up;
  vector<int> cyl_y_down;
  turn(cyl_z_up,cyl_x_up,1);
  turn(cyl_z_down,cyl_x_down,1);
  turn(cyl_z_up,cyl_y_up,-1);
  turn(cyl_z_down,cyl_y_down,-1);
    
  for(int i=0; i<3*M; i++) {
    vector<int> v;
    v.push_back(cyl_x_up[i]);
    v.push_back(cyl_x_down[i]);
    v.push_back(cyl_x_down[i+1]);
    v.push_back(cyl_x_up[i+1]);
    faces.push_back(v);
  }
  for(int i=0; i<3*M; i++) {
    vector<int> v;
    v.push_back(cyl_y_up[i]);
    v.push_back(cyl_y_down[i]);
    v.push_back(cyl_y_down[i+1]);
    v.push_back(cyl_y_up[i+1]);
    faces.push_back(v);
  }
  for(int i=0; i<3*M; i++) {
    vector<int> v;
    v.push_back(cyl_z_up[i]);
    v.push_back(cyl_z_down[i]);
    v.push_back(cyl_z_down[i+1]);
    v.push_back(cyl_z_up[i+1]);
    faces.push_back(v);
  }
  
  merg[cyl_z_up[0]]=cyl_y_down[3*M];
  merg[cyl_x_up[0]]=cyl_z_down[3*M];
  merg[cyl_y_up[0]]=cyl_x_down[3*M];

//  cout << p.x << " " << p.y << " " << p.z << endl;
  
  // Au tour des tores
  
  // le plus gros : tor,
  // R = 2, r = 1, outer angle span = pi/2, inner angle span = 3pi/2 
  
  vector<int> tor_z;
  for(int i=0; i<=2*M; i++) {
    th=(M_PI/2)*i/(double) M;
    for(int j=0; j<=3*M; j++) {
      phi=-M_PI/2+(M_PI/2)*(j/(double) M);
      add_point(1-(2+cos(phi))*cos(th),-1-(2+cos(phi))*sin(th),1+sin(phi)
               ,-cos(phi)*cos(th),-cos(phi)*sin(th),sin(phi));
      mark(tor_z);
    }
  }
  vector<int> tor_x;
  vector<int> tor_y;
  turn(tor_z,tor_x,1);
  turn(tor_z,tor_y,-1);
  
  for(int i=0; i<=2*M; i++) {
    merg[tor_z[i*(3*M+1)+3*M]]=cyl_z_up[i];
    merg[tor_x[i*(3*M+1)+3*M]]=cyl_x_up[i];
    merg[tor_y[i*(3*M+1)+3*M]]=cyl_y_up[i];
  }

  for(int j=0; j<=3*M; j++) {
    merg[tor_z[0*(3*M+1)+j]]=cyl_y_down[j];
    merg[tor_x[0*(3*M+1)+j]]=cyl_z_down[j];
    merg[tor_y[0*(3*M+1)+j]]=cyl_x_down[j];
  }
  
  vector<int> tor_x_faces;
  vector<int> tor_y_faces;
  vector<int> tor_z_faces;
  push_array(tor_x, 2*M, 3*M, &tor_x_faces);
  push_array(tor_y, 2*M, 3*M, &tor_y_faces);
  push_array(tor_z, 2*M, 3*M, &tor_z_faces);
  
  // torB
  // R=4, r=1, angle ext = pi/2-alpha2, angle int = alpha3
    
  double alpha1=atan(0.5);
  double alpha2=2*alpha1;
  double alpha3=(M_PI+alpha2)/2.;
  double alpha4=M_PI/2.-alpha2;
  
  // torB: vertex creation
  
  vector<int> torB_z;
  
  int N2=1+floor(alpha3/(M_PI/2)*M);
  for(int i=0; i<=N1; i++) {
    th=M_PI+alpha4*i/(double) N1;
    for(int j=0; j<=N2; j++) {
      if(j==N2)
        phi=-M_PI/2.+alpha3;
      else
        phi=-M_PI/2.+M_PI/2*(j/(double) M);
      add_point(-1-(4+cos(phi))*cos(th),-1-(4+cos(phi))*sin(th),1+sin(phi)
               ,-cos(phi)*cos(th),-cos(phi)*sin(th),sin(phi));
      mark(torB_z);
    }
  }
  vector<int> torB_x;
  vector<int> torB_y;
  turn(torB_z,torB_x,1);
  turn(torB_z,torB_y,-1);
  
  // torB: faces creation
  
  push_array(torB_x, N1, N2);
  push_array(torB_y, N1, N2);
  push_array(torB_z, N1, N2);
  
  // torB: vertices merger
  
  for(int j=0; j<N2; j++) {
    merg[torB_z[0*(N2+1)+j]]=tor_z[2*M*(3*M+1)+j];
    merg[torB_x[0*(N2+1)+j]]=tor_x[2*M*(3*M+1)+j];
    merg[torB_y[0*(N2+1)+j]]=tor_y[2*M*(3*M+1)+j];
  } 
    
  // torC
  // R=2 r=1 a-ext=alpha3 a-int=pi/2
  
  vector<int> torC_z;
  int N3=1+floor(alpha4/(M_PI/2.)*M);
  for(int i=0; i<=N3; i++) {
    if(i==N3) th=-alpha2;
    else th=-M_PI/2.+M_PI/2.*i/(double) M;
    //th=-M_PI/2.+(M_PI*1.5-alpha2)/2.*(i/(double) N3);
    for(int j=0; j<=M; j++) {
      phi=M_PI/2.+M_PI/2.*j/(double) M;
      add_point(1+(2+cos(phi))*cos(th),3+(2+cos(phi))*sin(th),-1+sin(phi)
               ,cos(phi)*cos(th),cos(phi)*sin(th),sin(phi));
      mark(torC_z);
    }
  }
  vector<int> torC_x;
  vector<int> torC_y;
  turn(torC_z,torC_x,1);
  turn(torC_z,torC_y,-1);
  
  push_array(torC_x, N3, M);
  push_array(torC_y, N3, M);
  push_array(torC_z, N3, M);
  
  // torC: merging vertices
  
  for(int j=0; j<=M; j++) {
    merg[torC_z[0*(M+1)+j]]=cyl_x_up[3*M-j];
    merg[torC_x[0*(M+1)+j]]=cyl_y_up[3*M-j];
    merg[torC_y[0*(M+1)+j]]=cyl_z_up[3*M-j];
  }
  merg[torC_z[N3*(M+1)+0]]=torB_z[N1*(N2+1)+0];
  merg[torC_x[N3*(M+1)+0]]=torB_x[N1*(N2+1)+0];
  merg[torC_y[N3*(M+1)+0]]=torB_y[N1*(N2+1)+0];
  for(int i=0; i<N3; i++) {
    merg[torC_z[i*(M+1)+M]]=tor_x[2*M*(3*M+1)+3*M-i];
    merg[torC_x[i*(M+1)+M]]=tor_y[2*M*(3*M+1)+3*M-i];
    merg[torC_y[i*(M+1)+M]]=tor_z[2*M*(3*M+1)+3*M-i];
  }
  
  // torD
  // R=2, r=1, ext-angle=pi/2, int_angle=alpha3
  
  vector<int> torD_z;
  double co=cos(alpha4);
  double si=sin(alpha4);
  for(int i=0; i<=M; i++) {
    point p;
    th=M_PI/2.*i/(double) M;
    for(int j=0; j<=N2; j++) {
      double x= (j==N2) ? 1 : M_PI/2*(j/(double) M)/alpha3;
      phi=M_PI-x*alpha3;
      p=point(sin(phi),(2+cos(phi))*cos(th),(2+cos(phi))*sin(th)
             ,sin(phi),cos(phi)*cos(th),cos(phi)*sin(th));
      p=p+point(0,-2,0);
      p=turn_z(p,co,si);
      p=p+point(1,3,-1);
      add_point(p);
      mark(torD_z);
    }
  }
  vector<int> torD_x;
  vector<int> torD_y;
  turn(torD_z,torD_x,1);
  turn(torD_z,torD_y,-1);
  
  vector<int> torD_x_faces;
  vector<int> torD_y_faces;
  vector<int> torD_z_faces;
  push_array(torD_x, M, N2, &torD_x_faces);
  push_array(torD_y, M, N2, &torD_y_faces);
  push_array(torD_z, M, N2, &torD_z_faces);
  
  // torD: vertices mergers
  
  for(int k=0; k<=M; k++) {
    merg[torD_z[k*(N2+1)+0]]=torC_z[N3*(M+1)+M-k];
    merg[torD_x[k*(N2+1)+0]]=torC_x[N3*(M+1)+M-k];
    merg[torD_y[k*(N2+1)+0]]=torC_y[N3*(M+1)+M-k];
  }
  for(int j=0; j<=N2; j++) {
    merg[torD_z[M*(N2+1)+j]]=torB_z[N1*(N2+1)+j];
    merg[torD_x[M*(N2+1)+j]]=torB_x[N1*(N2+1)+j];
    merg[torD_y[M*(N2+1)+j]]=torB_y[N1*(N2+1)+j];
  }
  int K0=floor(alpha4*M/(M_PI/2.));
  
  // torD: modifying faces on one side
  // quadri -> penta
  for(int j=0; j<N2; j++) {
    int jj=(int)floor(3*M-2*M*alpha4/M_PI-j);
    if(j==N2-1 && -M_PI/2.+M_PI/(2*M)*jj < M_PI-alpha4-alpha3)
        break;
    vector<int> v;
    v.push_back(torD_z[0*(N2+1)+j]);
    v.push_back(tor_x[2*M*(3*M+1)+jj]);
    v.push_back(torD_z[0*(N2+1)+j+1]);
    v.push_back(torD_z[1*(N2+1)+j+1]);
    v.push_back(torD_z[1*(N2+1)+j]);
    faces[torD_z_faces[0*N2+j]]=v;
    v.clear();
    v.push_back(torD_x[0*(N2+1)+j]);
    v.push_back(tor_y[2*M*(3*M+1)+jj]);
    v.push_back(torD_x[0*(N2+1)+j+1]);
    v.push_back(torD_x[1*(N2+1)+j+1]);
    v.push_back(torD_x[1*(N2+1)+j]);
    faces[torD_x_faces[0*N2+j]]=v;
    v.clear();
    v.push_back(torD_y[0*(N2+1)+j]);
    v.push_back(tor_z[2*M*(3*M+1)+jj]);
    v.push_back(torD_y[0*(N2+1)+j+1]);
    v.push_back(torD_y[1*(N2+1)+j+1]);
    v.push_back(torD_y[1*(N2+1)+j]);
    faces[torD_y_faces[0*N2+j]]=v;
  }
  
  // tor: modif faces
  // quadri -> penta
  int jj0=(int)floor(3*M-2*M*alpha4/M_PI-(N2-1));
  bool tst = M_PI-alpha4-alpha3>-M_PI/2.+M_PI/(2*M)*jj0;
  for(int j=0; j<=N2; j++){
    vector<int> v;
    int jj=(int)floor(3*M-2*M*alpha4/M_PI-j);
    if(j==N2 && tst)
      break;
    v.push_back(tor_x[(2*M-1)*(3*M+1)+jj]);
    v.push_back(tor_x[    2*M*(3*M+1)+jj]);
    if(jj==N2-1 && tst) 
      v.push_back(torD_z[0*(N2+1)+N2]);
    v.push_back(torD_z[0*(N2+1)+j]);
    v.push_back(tor_x[    2*M*(3*M+1)+jj+1]);
    v.push_back(tor_x[(2*M-1)*(3*M+1)+jj+1]);
    faces[tor_x_faces[(2*M-1)*3*M+jj]]=v;
    v.clear();
    v.push_back(tor_y[(2*M-1)*(3*M+1)+jj]);
    v.push_back(tor_y[    2*M*(3*M+1)+jj]);
    if(jj==N2-1 && tst) 
      v.push_back(torD_x[0*(N2+1)+N2]);
    v.push_back(torD_x[0*(N2+1)+j]);
    v.push_back(tor_y[    2*M*(3*M+1)+jj+1]);
    v.push_back(tor_y[(2*M-1)*(3*M+1)+jj+1]);
    faces[tor_y_faces[(2*M-1)*3*M+jj]]=v;
    v.clear();
    v.push_back(tor_z[(2*M-1)*(3*M+1)+jj]);
    v.push_back(tor_z[    2*M*(3*M+1)+jj]);
    if(jj==N2-1 && tst) 
      v.push_back(torD_y[0*(N2+1)+N2]);
    v.push_back(torD_y[0*(N2+1)+j]);
    v.push_back(tor_z[    2*M*(3*M+1)+jj+1]);
    v.push_back(tor_z[(2*M-1)*(3*M+1)+jj+1]);
    faces[tor_z_faces[(2*M-1)*3*M+jj]]=v;    
  } 

  merg[torD_z[0*(N2+1)+N2]]=torB_x[0*(N2+1)+N2];
  merg[torD_x[0*(N2+1)+N2]]=torB_y[0*(N2+1)+N2];
  merg[torD_y[0*(N2+1)+N2]]=torB_z[0*(N2+1)+N2]; 

  // Puis les plats
  
  vector<int> flat_z;
  vector<int> flat_x;
  vector<int> flat_y;
  
  // aucune création de vertex !
  
  flat_z.push_back(cyl_y_up[0]);
  flat_z.push_back(cyl_y_down[0]);
  for(int i=1; i<=2*M; i++) {
    flat_z.push_back(tor_z[i*(3*M+1)+0]);
  }
  for(int i=1; i<N1; i++) {
    flat_z.push_back(torB_z[i*(N2+1)+0]);
  }  
  for(int i=0; i<N3; i++) {
    flat_z.push_back(torC_z[(N3-i)*(M+1)+0]);
  }  
  flat_z.push_back(cyl_x_up[3*M]);

  flat_x.push_back(cyl_z_up[0]);
  flat_x.push_back(cyl_z_down[0]);
  for(int i=1; i<=2*M; i++) {
    flat_x.push_back(tor_x[i*(3*M+1)+0]);
  }
  for(int i=1; i<N1; i++) {
    flat_x.push_back(torB_x[i*(N2+1)+0]);
  }  
  for(int i=0; i<N3; i++) {
    flat_x.push_back(torC_x[(N3-i)*(M+1)+0]);
  }  
  flat_x.push_back(cyl_y_up[3*M]);  

  flat_y.push_back(cyl_x_up[0]);
  flat_y.push_back(cyl_x_down[0]);
  for(int i=1; i<=2*M; i++) {
    flat_y.push_back(tor_y[i*(3*M+1)+0]);
  }
  for(int i=1; i<N1; i++) {
    flat_y.push_back(torB_y[i*(N2+1)+0]);
  }  
  for(int i=0; i<N3; i++) {
    flat_y.push_back(torC_y[(N3-i)*(M+1)+0]);
  }  
  flat_y.push_back(cyl_z_up[3*M]);  
  
  {
    vector<int> v;
    for(int i=0; i<flat_x.size(); i++) {
      v.push_back(flat_x[i]);
    }
    faces.push_back(v);
    v.clear();
    for(int i=0; i<flat_y.size(); i++) {
      v.push_back(flat_y[i]);
    }
    faces.push_back(v);
    v.clear();
    for(int i=0; i<flat_z.size(); i++) {
      v.push_back(flat_z[i]);
    }
    faces.push_back(v);
  }
  
  // Enfin des bouts de la sphère
  
  vector<int> sph1;
  double rad=2.*sqrt(5.)+1.;
  for(int i=0; i<=N1; i++) {
    point p,p1,p0;
    point off(1,1,1);
    p0=vertices[torB_z[i*(N2+1)+N2]];
    p1=vertices[torD_y[i*(N2+1)+N2]];
    for(int j=0; j<=i; j++) {
      if(j==0) p=p0;
      else {
        p=(p0*(i-j)+p1*j)*(1./i);
      }
      p=p+off;
      p.nx=p.x;
      p.ny=p.y;
      p.nz=p.z;
      p=p*(1./norm(p));
      p.x *= rad;
      p.y *= rad;
      p.z *= rad;
      p=p+(-1.)*off;
      add_point(p);
      mark(sph1);
    }    
  }
  vector<int> sph2;
  for(int i=0; i<=N1; i++) {
    point p,p1,p0;
    point off(1,1,1);
    p0=vertices[torB_x[i*(N2+1)+N2]];
    p1=vertices[torD_z[i*(N2+1)+N2]];
    for(int j=0; j<=i; j++) {
      if(j==0) p=p0;
      else {
        p=(p0*(i-j)+p1*j)*(1./i);
      }
      p=p+off;
      p.nx=p.x;
      p.ny=p.y;
      p.nz=p.z;
      p=p*(1./norm(p));
      p.x *= rad;
      p.y *= rad;
      p.z *= rad;
      p=p+(-1.)*off;
      add_point(p);
      mark(sph2);
    }    
  }
  vector<int> sph3;
  for(int i=0; i<=N1; i++) {
    point p,p1,p0;
    point off(1,1,1);
    p0=vertices[torB_y[i*(N2+1)+N2]];
    p1=vertices[torD_x[i*(N2+1)+N2]];
    for(int j=0; j<=i; j++) {
      if(j==0) p=p0;
      else {
        p=(p0*(i-j)+p1*j)*(1./i);
      }
      p=p+off;
      p.nx=p.x;
      p.ny=p.y;
      p.nz=p.z;
      p=p*(1./norm(p));
      p.x *= rad;
      p.y *= rad;
      p.z *= rad;
      p=p+(-1.)*off;
      add_point(p);
      mark(sph3);
    }    
  }
  vector<int> sph4;
  for(int i=0; i<=N1; i++) {
    point p,p1,p0;
    point off(1,1,1);
    int n=(N1*(N1+1))/2;
    p0=vertices[sph2[n+N1-i]];
    p1=vertices[sph1[n+i]];
    for(int j=0; j<=i; j++) {
      if(j==0) p=p0;
      else {
        p=(p0*(i-j)+p1*j)*(1./i);
      }
      p=p+off;
      p.nx=p.x;
      p.ny=p.y;
      p.nz=p.z;
      p=p*(1./norm(p));
      p.x *= rad;
      p.y *= rad;
      p.z *= rad;
      p=p+(-1.)*off;
      add_point(p);
      mark(sph4);
    }    
  }
  
  {
    vector<int> v;
    for(int i=0; i<N1; i++) {
      for(int j=0; j<=i; j++) {
        v.clear();
        int n=(i*(i+1))/2;
        v.push_back(sph1[n+j]);
        v.push_back(sph1[n+i+1+j]);
        v.push_back(sph1[n+i+1+j+1]);
        faces.push_back(v);
      }
      for(int j=0; j<i; j++) {
        v.clear();
        int n=(i*(i+1))/2;
        v.push_back(sph1[n+j+1]);
        v.push_back(sph1[n+j]);
        v.push_back(sph1[n+i+1+j+1]);
        faces.push_back(v);
      }
      for(int j=0; j<=i; j++) {
        v.clear();
        int n=(i*(i+1))/2;
        v.push_back(sph2[n+j]);
        v.push_back(sph2[n+i+1+j]);
        v.push_back(sph2[n+i+1+j+1]);
        faces.push_back(v);
      }
      for(int j=0; j<i; j++) {
        v.clear();
        int n=(i*(i+1))/2;
        v.push_back(sph2[n+j+1]);
        v.push_back(sph2[n+j]);
        v.push_back(sph2[n+i+1+j+1]);
        faces.push_back(v);
      }
      for(int j=0; j<=i; j++) {
        v.clear();
        int n=(i*(i+1))/2;
        v.push_back(sph3[n+j]);
        v.push_back(sph3[n+i+1+j]);
        v.push_back(sph3[n+i+1+j+1]);
        faces.push_back(v);
      }
      for(int j=0; j<i; j++) {
        v.clear();
        int n=(i*(i+1))/2;
        v.push_back(sph3[n+j+1]);
        v.push_back(sph3[n+j]);
        v.push_back(sph3[n+i+1+j+1]);
        faces.push_back(v);
      }
      for(int j=0; j<=i; j++) {
        v.clear();
        int n=(i*(i+1))/2;
        v.push_back(sph4[n+j]);
        v.push_back(sph4[n+i+1+j]);
        v.push_back(sph4[n+i+1+j+1]);
        faces.push_back(v);
      }
      for(int j=0; j<i; j++) {
        v.clear();
        int n=(i*(i+1))/2;
        v.push_back(sph4[n+j+1]);
        v.push_back(sph4[n+j]);
        v.push_back(sph4[n+i+1+j+1]);
        faces.push_back(v);
      }
    }
  }
  
  for(int i=0; i<=N1; i++) {
    int n=(i*(i+1))/2;
    int NN=(N1*(N1+1))/2;
    merg[sph1[n]]=torB_z[i*(N2+1)+N2];
    merg[sph1[n+i]]=torD_y[i*(N2+1)+N2];
    merg[sph2[n]]=torB_x[i*(N2+1)+N2];
    merg[sph2[n+i]]=torD_z[i*(N2+1)+N2];
    merg[sph3[n]]=torB_y[i*(N2+1)+N2];
    merg[sph3[n+i]]=torD_x[i*(N2+1)+N2];
    merg[sph4[NN+i]]=sph3[NN+N1-i];
    merg[sph4[(i*(i+1))/2+i]]=sph1[NN+i];
    merg[sph4[(i*(i+1))/2]]=sph2[NN+N1-i];
  }
  
}

void do_mergers() {
  vector<point> verts; // array of new points
  vector<int> o_n; // old index -> index of the corresponding point in the new array
  vector<int> n_o; // new index -> index of the corresponding point in the old array
  int new_index=-1;
  for(int old_index=0; old_index<vertices.size(); old_index++) {
    if(merg[old_index]==-1) {
      new_index++;
      verts.push_back(vertices[old_index]);
      n_o.push_back(old_index);
      o_n.push_back(new_index);
    }
    else o_n.push_back(-1);
  }
  for(int old_index=0; old_index<vertices.size(); old_index++) {
    if(merg[old_index]!=-1) {
      int k=old_index;
      //cout << k;
      while(merg[k]!=-1) {     // !!! may get stuck in an infinite loop
        k=merg[k];             // optimizable btw
        //cout << " -> " << k;
      }
      //cout << endl;
      o_n[old_index]=o_n[k];
    }
  }
  vertices=verts; // inefficient, but not critical

  for(int n=0; n<faces.size(); n++) {
    vector<int> &v=faces[n];
    for(int k=0; k<v.size(); k++) { 
      v[k]=o_n[v[k]];
    }
  }
  
}

void thicken() {
  vector<point> vs;
  vector<vector<int> > fs;
  point p,p0;
  for(int i=0; i<vertices.size(); i++) {
    p0=vertices[i];
    double d=norm(p0.normal());
    if(d==0) cout << "WARNING: normal[" << i << "]=0" << endl;
    p0.nx /= d;
    p0.ny /= d;
    p0.nz /= d;
    p=p0+h*p0.normal();
    vs.push_back(p);
    p=p0+(-h)*p0.normal();
    p.nx=-p.nx;
    p.ny=-p.ny;
    p.nz=-p.nz;
    vs.push_back(p);
  } 
  for(int i=0; i<faces.size(); i++) {
    vector<int> f=faces[i];
    
    // compute barycenter of the face
    point p; // 0 by default
    for(int k=0; k<f.size(); k++) {
      p=p+vertices[f[k]];
    }
    p=p*(1./(double) f.size());
    // compute approx normal vector for the face
    point nor=cross(vertices[f[0]]+(-1.)*p,vertices[f[1]]+(-1.)*p);
    
    vector<int> v;    
    for(int k=0; k<f.size(); k++) {      
      if(dot(nor,vertices[f[k]].normal())>0)
        v.push_back(2*f[k]);
      else
        v.push_back(2*f[k]+1);
    }
    fs.push_back(v);
    
    vector<int> w;
    for(int k=f.size()-1; k>=0; k--) {      
      if(dot(nor,vertices[f[k]].normal())>0)
        w.push_back(2*f[k]+1);
      else
        w.push_back(2*f[k]);
    }
    fs.push_back(w);
  }
  vertices=vs; // deep copy: inefficient but it does not matter here
  faces=fs; // idem 
}

string s;

void fill_string() {
  char aux[100];
  for(vector<point>::iterator it=vertices.begin(); it<vertices.end(); it++) {
    point &v=*it;
    s.append("v");
    sprintf(aux," %f",v.x);
    s.append(aux);
    sprintf(aux," %f",v.y);
    s.append(aux);
    sprintf(aux," %f\n",v.z);
    s.append(aux);
  }
  s.append("\n");
  for(vector<point>::iterator it=vertices.begin(); it<vertices.end(); it++) {
    point &v=*it;
    s.append("vn");
    sprintf(aux," %f",v.nx);
    s.append(aux);
    sprintf(aux," %f",v.ny);
    s.append(aux);
    sprintf(aux," %f\n",v.nz);
    s.append(aux);
  }
  s.append("\n");
  for(vector<vector<int> >::iterator it=faces.begin(); it<faces.end(); it++) {
    s.append("f");
    vector<int> &f=*it;
    for(vector<int>::iterator id=f.begin(); id<f.end(); id++) {
      sprintf(aux," %i",1+(*id));
      s.append(aux);
    }
    s.append("\n");
  }
}

void write_string(const char* filename) {
  FILE *file=fopen(filename,"wb");
      if(file==NULL) { cout << "could not open file" << endl; exit(-1); }
  int err=fputs(s.c_str(),file);
      if(err==EOF) { cout << "could not write on file" << endl; exit(-1); }
  err=fclose(file);
      if(err==EOF) { cout << "could not close file" << endl; exit(-1); }
}

int main(int argc, char **argv) {

  // create data structures
  build_surface();
   
  //
  cout << faces.size() << " faces" << endl;
  cout << vertices.size() << " vertices before mergers" << endl;

  do_mergers();

  cout << vertices.size() << " vertices after mergers" << endl;

  // transform non-orientable flat surface into its orientable 2-cover with some shift,
  // so that it encloses a volume
  
  thicken();

  cout << vertices.size() << " vertices after doubling" << endl;
  cout << faces.size() << " faces after doubling" << endl;
 
   // put all data in string
  s="";
  fill_string();
  
  // put string in file
  write_string("mesh.obj");
}